@@ -1,26 +1,26 @@
 /*****************************************************************************
  * stream_output.c : stream output module
  *****************************************************************************
- * Copyright (C) 2002-2007 the VideoLAN team
- * $Id: 5777fbd44c8ca4be9013bcdfa269fde28f5f8eb5 $
+ * Copyright (C) 2002-2007 VLC authors and VideoLAN
+ * $Id: 35a51f4f3c35e08c1bdd9a06c5d2e0d4fd60aa38 $
  *
  * Authors: Christophe Massiot <massiot@via.ecp.fr>
  *          Laurent Aimar <fenrir@via.ecp.fr>
  *          Eric Petit <titer@videolan.org>
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * GNU Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
  *****************************************************************************/
 
 /*****************************************************************************
@@ -31,6 +31,8 @@
 # include "config.h"
 #endif
 
+#include <assert.h>
+
 #include <vlc_common.h>
 
 #include <stdlib.h>                                                /* free() */
@@ -44,6 +46,7 @@
 #include <vlc_meta.h>
 #include <vlc_block.h>
 #include <vlc_codec.h>
+#include <vlc_modules.h>
 
 #include "input/input_interface.h"
 
@@ -51,9 +54,7 @@
 /*****************************************************************************
  * Local prototypes
  *****************************************************************************/
-#define sout_stream_url_to_chain( p, s ) \
-    _sout_stream_url_to_chain( VLC_OBJECT(p), s )
-static char *_sout_stream_url_to_chain( vlc_object_t *, const char * );
+static char *sout_stream_url_to_chain( bool, const char * );
 
 /*
  * Generic MRL parser
@@ -72,59 +73,64 @@ static int  mrl_Parse( mrl_t *p_mrl, const char *psz_mrl );
 /* mrl_Clean: clean p_mrl  after a call to mrl_Parse */
 static void mrl_Clean( mrl_t *p_mrl );
 
+#undef sout_NewInstance
+
 /*****************************************************************************
  * sout_NewInstance: creates a new stream output instance
  *****************************************************************************/
-sout_instance_t *__sout_NewInstance( vlc_object_t *p_parent, const char *psz_dest )
+sout_instance_t *sout_NewInstance( vlc_object_t *p_parent, const char *psz_dest )
 {
-    static const char typename[] = "stream output";
     sout_instance_t *p_sout;
+    char *psz_chain;
 
-    /* *** Allocate descriptor *** */
-    p_sout = vlc_custom_create( p_parent, sizeof( *p_sout ),
-                                VLC_OBJECT_GENERIC, typename );
-    if( p_sout == NULL )
-        return NULL;
+    assert( psz_dest != NULL );
 
-    /* *** init descriptor *** */
-    p_sout->psz_sout    = strdup( psz_dest );
-    p_sout->p_meta      = NULL;
-    p_sout->i_out_pace_nocontrol = 0;
-    p_sout->p_sys       = NULL;
-
-    vlc_mutex_init( &p_sout->lock );
-    if( psz_dest && psz_dest[0] == '#' )
+    if( psz_dest[0] == '#' )
     {
-        p_sout->psz_chain = strdup( &psz_dest[1] );
+        psz_chain = strdup( &psz_dest[1] );
     }
     else
     {
-        p_sout->psz_chain = sout_stream_url_to_chain( p_sout, psz_dest );
-        msg_Dbg( p_sout, "using sout chain=`%s'", p_sout->psz_chain );
+        psz_chain = sout_stream_url_to_chain(
+            var_InheritBool(p_parent, "sout-display"), psz_dest );
     }
-    p_sout->p_stream = NULL;
+    if(!psz_chain)
+        return NULL;
 
-    /* attach it for inherit */
-    vlc_object_attach( p_sout, p_parent );
-
-    /* */
-    var_Create( p_sout, "sout-mux-caching", VLC_VAR_INTEGER | VLC_VAR_DOINHERIT );
-
-    /* */
-    p_sout->p_stream = sout_StreamNew( p_sout, p_sout->psz_chain );
-    if( p_sout->p_stream == NULL )
+    /* *** Allocate descriptor *** */
+    p_sout = vlc_custom_create( p_parent, sizeof( *p_sout ), "stream output" );
+    if( p_sout == NULL )
     {
-        msg_Err( p_sout, "stream chain failed for `%s'", p_sout->psz_chain );
-
-        FREENULL( p_sout->psz_sout );
-        FREENULL( p_sout->psz_chain );
-
-        vlc_object_detach( p_sout );
-        vlc_object_release( p_sout );
+        free( psz_chain );
         return NULL;
     }
 
-    return p_sout;
+    msg_Dbg( p_sout, "using sout chain=`%s'", psz_chain );
+
+    /* *** init descriptor *** */
+    p_sout->psz_sout    = strdup( psz_dest );
+    p_sout->i_out_pace_nocontrol = 0;
+
+    vlc_mutex_init( &p_sout->lock );
+    p_sout->p_stream = NULL;
+
+    var_Create( p_sout, "sout-mux-caching", VLC_VAR_INTEGER | VLC_VAR_DOINHERIT );
+
+    p_sout->p_stream = sout_StreamChainNew( p_sout, psz_chain, NULL, NULL );
+    if( p_sout->p_stream )
+    {
+        free( psz_chain );
+        return p_sout;
+    }
+
+    msg_Err( p_sout, "stream chain failed for `%s'", psz_chain );
+    free( psz_chain );
+
+    FREENULL( p_sout->psz_sout );
+
+    vlc_mutex_destroy( &p_sout->lock );
+    vlc_object_release( p_sout );
+    return NULL;
 }
 
 /*****************************************************************************
@@ -133,17 +139,10 @@ sout_instance_t *__sout_NewInstance( vlc_object_t *p_parent, const char *psz_des
 void sout_DeleteInstance( sout_instance_t * p_sout )
 {
     /* remove the stream out chain */
-    sout_StreamDelete( p_sout->p_stream );
+    sout_StreamChainDelete( p_sout->p_stream, NULL );
 
     /* *** free all string *** */
     FREENULL( p_sout->psz_sout );
-    FREENULL( p_sout->psz_chain );
-
-    /* delete meta */
-    if( p_sout->p_meta )
-    {
-        vlc_meta_Delete( p_sout->p_meta );
-    }
 
     vlc_mutex_destroy( &p_sout->lock );
 
@@ -151,71 +150,22 @@ void sout_DeleteInstance( sout_instance_t * p_sout )
     vlc_object_release( p_sout );
 }
 
-/*****************************************************************************
- * 
- *****************************************************************************/
-void sout_UpdateStatistic( sout_instance_t *p_sout, sout_statistic_t i_type, int i_delta )
-{
-    if( !libvlc_stats( p_sout ) )
-        return;
-
-    /* */
-    input_thread_t *p_input = vlc_object_find( p_sout, VLC_OBJECT_INPUT, FIND_PARENT );
-    if( !p_input )
-        return;
-
-    int i_input_type;
-    switch( i_type )
-    {
-    case SOUT_STATISTIC_DECODED_VIDEO:
-        i_input_type = SOUT_STATISTIC_DECODED_VIDEO;
-        break;
-    case SOUT_STATISTIC_DECODED_AUDIO:
-        i_input_type = SOUT_STATISTIC_DECODED_AUDIO;
-        break;
-    case SOUT_STATISTIC_DECODED_SUBTITLE:
-        i_input_type = SOUT_STATISTIC_DECODED_SUBTITLE;
-        break;
-
-    case SOUT_STATISTIC_SENT_PACKET:
-        i_input_type = SOUT_STATISTIC_SENT_PACKET;
-        break;
-
-    case SOUT_STATISTIC_SENT_BYTE:
-        i_input_type = SOUT_STATISTIC_SENT_BYTE;
-        break;
-
-    default:
-        msg_Err( p_sout, "Not yet supported statistic type %d", i_type );
-        vlc_object_release( p_input );
-        return;
-    }
-
-    input_UpdateStatistic( p_input, i_input_type, i_delta );
-
-    vlc_object_release( p_input );
-}
 /*****************************************************************************
  * Packetizer/Input
  *****************************************************************************/
 sout_packetizer_input_t *sout_InputNew( sout_instance_t *p_sout,
-                                        es_format_t *p_fmt )
+                                        const es_format_t *p_fmt )
 {
     sout_packetizer_input_t *p_input;
 
     /* *** create a packetizer input *** */
-    p_input         = malloc( sizeof( sout_packetizer_input_t ) );
-    if( !p_input )  return NULL;
+    if( !p_fmt->i_codec || !(p_input = malloc(sizeof(sout_packetizer_input_t))) )
+        return NULL;
+
     p_input->p_sout = p_sout;
-    p_input->p_fmt  = p_fmt;
 
-    msg_Dbg( p_sout, "adding a new sout input (sout_input:%p)", p_input );
-
-    if( p_fmt->i_codec == VLC_FOURCC( 'n', 'u', 'l', 'l' ) )
-    {
-        vlc_object_release( p_sout );
-        return p_input;
-    }
+    msg_Dbg( p_sout, "adding a new sout input for `%4.4s` (sout_input: %p)",
+             (char*) &p_fmt->i_codec, (void *)p_input );
 
     /* *** add it to the stream chain */
     vlc_mutex_lock( &p_sout->lock );
@@ -224,8 +174,10 @@ sout_packetizer_input_t *sout_InputNew( sout_instance_t *p_sout,
 
     if( p_input->id == NULL )
     {
+        msg_Warn( p_sout, "new sout input failed (sout_input: %p)",
+                 (void *)p_input );
         free( p_input );
-        return NULL;
+        p_input = NULL;
     }
 
     return( p_input );
@@ -238,20 +190,39 @@ int sout_InputDelete( sout_packetizer_input_t *p_input )
 {
     sout_instance_t     *p_sout = p_input->p_sout;
 
-    msg_Dbg( p_sout, "removing a sout input (sout_input:%p)", p_input );
+    msg_Dbg( p_sout, "removing a sout input (sout_input: %p)",
+             (void *)p_input );
 
-    if( p_input->p_fmt->i_codec != VLC_FOURCC( 'n', 'u', 'l', 'l' ) )
-    {
-        vlc_mutex_lock( &p_sout->lock );
-        p_sout->p_stream->pf_del( p_sout->p_stream, p_input->id );
-        vlc_mutex_unlock( &p_sout->lock );
-    }
+    vlc_mutex_lock( &p_sout->lock );
+    p_sout->p_stream->pf_del( p_sout->p_stream, p_input->id );
+    vlc_mutex_unlock( &p_sout->lock );
 
     free( p_input );
 
     return( VLC_SUCCESS);
 }
 
+bool sout_InputIsEmpty( sout_packetizer_input_t *p_input )
+{
+    sout_instance_t *p_sout = p_input->p_sout;
+    bool b;
+
+    vlc_mutex_lock( &p_sout->lock );
+    if( sout_StreamControl( p_sout->p_stream, SOUT_STREAM_EMPTY, &b ) != VLC_SUCCESS )
+        b = true;
+    vlc_mutex_unlock( &p_sout->lock );
+    return b;
+}
+
+void sout_InputFlush( sout_packetizer_input_t *p_input )
+{
+    sout_instance_t     *p_sout = p_input->p_sout;
+
+    vlc_mutex_lock( &p_sout->lock );
+    sout_StreamFlush( p_sout->p_stream, p_input->id );
+    vlc_mutex_unlock( &p_sout->lock );
+}
+
 /*****************************************************************************
  *
  *****************************************************************************/
@@ -261,18 +232,6 @@ int sout_InputSendBuffer( sout_packetizer_input_t *p_input,
     sout_instance_t     *p_sout = p_input->p_sout;
     int                 i_ret;
 
-    if( p_input->p_fmt->i_codec == VLC_FOURCC( 'n', 'u', 'l', 'l' ) )
-    {
-        block_Release( p_buffer );
-        return VLC_SUCCESS;
-    }
-    if( p_buffer->i_dts <= 0 )
-    {
-        msg_Warn( p_sout, "trying to send non-dated packet to stream output!");
-        block_Release( p_buffer );
-        return VLC_SUCCESS;
-    }
-
     vlc_mutex_lock( &p_sout->lock );
     i_ret = p_sout->p_stream->pf_send( p_sout->p_stream,
                                        p_input->id, p_buffer );
@@ -288,12 +247,10 @@ int sout_InputSendBuffer( sout_packetizer_input_t *p_input,
 sout_access_out_t *sout_AccessOutNew( vlc_object_t *p_sout,
                                       const char *psz_access, const char *psz_name )
 {
-    static const char typename[] = "access out";
     sout_access_out_t *p_access;
     char              *psz_next;
 
-    p_access = vlc_custom_create( p_sout, sizeof( *p_access ),
-                                  VLC_OBJECT_GENERIC, typename );
+    p_access = vlc_custom_create( p_sout, sizeof( *p_access ), "access out" );
     if( !p_access )
         return NULL;
 
@@ -308,11 +265,6 @@ sout_access_out_t *sout_AccessOutNew( vlc_object_t *p_sout,
     p_access->pf_control = NULL;
     p_access->p_module   = NULL;
 
-    p_access->i_writes = 0;
-    p_access->i_sent_bytes = 0;
-
-    vlc_object_attach( p_access, p_sout );
-
     p_access->p_module   =
         module_need( p_access, "sout access", p_access->psz_access, true );
 
@@ -320,7 +272,6 @@ sout_access_out_t *sout_AccessOutNew( vlc_object_t *p_sout,
     {
         free( p_access->psz_access );
         free( p_access->psz_path );
-        vlc_object_detach( p_access );
         vlc_object_release( p_access );
         return( NULL );
     }
@@ -332,7 +283,6 @@ sout_access_out_t *sout_AccessOutNew( vlc_object_t *p_sout,
  *****************************************************************************/
 void sout_AccessOutDelete( sout_access_out_t *p_access )
 {
-    vlc_object_detach( p_access );
     if( p_access->p_module )
     {
         module_unneed( p_access, p_access->p_module );
@@ -351,6 +301,8 @@ void sout_AccessOutDelete( sout_access_out_t *p_access )
  *****************************************************************************/
 int sout_AccessOutSeek( sout_access_out_t *p_access, off_t i_pos )
 {
+    if (p_access->pf_seek == NULL)
+        return VLC_EGENERIC;
     return p_access->pf_seek( p_access, i_pos );
 }
 
@@ -368,17 +320,6 @@ ssize_t sout_AccessOutRead( sout_access_out_t *p_access, block_t *p_buffer )
  *****************************************************************************/
 ssize_t sout_AccessOutWrite( sout_access_out_t *p_access, block_t *p_buffer )
 {
-#if 0
-    const unsigned i_packets_gather = 30;
-    p_access->i_writes++;
-    p_access->i_sent_bytes += p_buffer->i_buffer;
-    if( (p_access->i_writes % i_packets_gather) == 0 )
-    {
-        sout_UpdateStatistic( p_access->p_sout, SOUT_STATISTIC_SENT_PACKET, i_packets_gather );
-        sout_UpdateStatistic( p_access->p_sout, SOUT_STATISTIC_SENT_BYTE, p_access->i_sent_bytes );
-        p_access->i_sent_bytes = 0;
-    }
-#endif
     return p_access->pf_write( p_access, p_buffer );
 }
 
@@ -402,15 +343,13 @@ int sout_AccessOutControl (sout_access_out_t *access, int query, ...)
 /*****************************************************************************
  * sout_MuxNew: create a new mux
  *****************************************************************************/
-sout_mux_t * sout_MuxNew( sout_instance_t *p_sout, char *psz_mux,
+sout_mux_t * sout_MuxNew( sout_instance_t *p_sout, const char *psz_mux,
                           sout_access_out_t *p_access )
 {
-    static const char typename[] = "mux";
     sout_mux_t *p_mux;
     char       *psz_next;
 
-    p_mux = vlc_custom_create( p_sout, sizeof( *p_mux ), VLC_OBJECT_GENERIC,
-                               typename);
+    p_mux = vlc_custom_create( p_sout, sizeof( *p_mux ), "mux" );
     if( p_mux == NULL )
         return NULL;
 
@@ -433,8 +372,6 @@ sout_mux_t * sout_MuxNew( sout_instance_t *p_sout, char *psz_mux,
     p_mux->b_waiting_stream = true;
     p_mux->i_add_stream_start = -1;
 
-    vlc_object_attach( p_mux, p_sout );
-
     p_mux->p_module =
         module_need( p_mux, "sout mux", p_mux->psz_mux, true );
 
@@ -442,7 +379,6 @@ sout_mux_t * sout_MuxNew( sout_instance_t *p_sout, char *psz_mux,
     {
         FREENULL( p_mux->psz_mux );
 
-        vlc_object_detach( p_mux );
         vlc_object_release( p_mux );
         return NULL;
     }
@@ -493,7 +429,6 @@ sout_mux_t * sout_MuxNew( sout_instance_t *p_sout, char *psz_mux,
  *****************************************************************************/
 void sout_MuxDelete( sout_mux_t *p_mux )
 {
-    vlc_object_detach( p_mux );
     if( p_mux->p_module )
     {
         module_unneed( p_mux, p_mux->p_module );
@@ -508,7 +443,7 @@ void sout_MuxDelete( sout_mux_t *p_mux )
 /*****************************************************************************
  * sout_MuxAddStream:
  *****************************************************************************/
-sout_input_t *sout_MuxAddStream( sout_mux_t *p_mux, es_format_t *p_fmt )
+sout_input_t *sout_MuxAddStream( sout_mux_t *p_mux, const es_format_t *p_fmt )
 {
     sout_input_t *p_input;
 
@@ -525,8 +460,11 @@ sout_input_t *sout_MuxAddStream( sout_mux_t *p_mux, es_format_t *p_fmt )
     p_input = malloc( sizeof( sout_input_t ) );
     if( !p_input )
         return NULL;
-    p_input->p_sout = p_mux->p_sout;
-    p_input->p_fmt  = p_fmt;
+
+    // FIXME: remove either fmt or p_fmt...
+    es_format_Copy( &p_input->fmt, p_fmt );
+    p_input->p_fmt = &p_input->fmt;
+
     p_input->p_fifo = block_FifoNew();
     p_input->p_sys  = NULL;
 
@@ -536,6 +474,7 @@ sout_input_t *sout_MuxAddStream( sout_mux_t *p_mux, es_format_t *p_fmt )
         msg_Err( p_mux, "cannot add this stream" );
         TAB_REMOVE( p_mux->i_nb_inputs, p_mux->pp_inputs, p_input );
         block_FifoRelease( p_input->p_fifo );
+        es_format_Clean( &p_input->fmt );
         free( p_input );
         return NULL;
     }
@@ -562,10 +501,7 @@ void sout_MuxDeleteStream( sout_mux_t *p_mux, sout_input_t *p_input )
     TAB_FIND( p_mux->i_nb_inputs, p_mux->pp_inputs, p_input, i_index );
     if( i_index >= 0 )
     {
-        if( p_mux->pf_delstream( p_mux, p_input ) < 0 )
-        {
-            msg_Err( p_mux, "cannot delete this stream from mux" );
-        }
+        p_mux->pf_delstream( p_mux, p_input );
 
         /* remove the entry */
         TAB_REMOVE( p_mux->i_nb_inputs, p_mux->pp_inputs, p_input );
@@ -576,6 +512,7 @@ void sout_MuxDeleteStream( sout_mux_t *p_mux, sout_input_t *p_input )
         }
 
         block_FifoRelease( p_input->p_fifo );
+        es_format_Clean( &p_input->fmt );
         free( p_input );
     }
 }
@@ -583,17 +520,18 @@ void sout_MuxDeleteStream( sout_mux_t *p_mux, sout_input_t *p_input )
 /*****************************************************************************
  * sout_MuxSendBuffer:
  *****************************************************************************/
-void sout_MuxSendBuffer( sout_mux_t *p_mux, sout_input_t *p_input,
+int sout_MuxSendBuffer( sout_mux_t *p_mux, sout_input_t *p_input,
                          block_t *p_buffer )
 {
+    vlc_tick_t i_dts = p_buffer->i_dts;
     block_FifoPut( p_input->p_fifo, p_buffer );
 
     if( p_mux->p_sout->i_out_pace_nocontrol )
     {
-        mtime_t current_date = mdate();
-        if ( current_date > p_buffer->i_dts )
+        vlc_tick_t current_date = mdate();
+        if ( current_date > i_dts )
             msg_Warn( p_mux, "late buffer for mux input (%"PRId64")",
-                      current_date - p_buffer->i_dts );
+                      current_date - i_dts );
     }
 
     if( p_mux->b_waiting_stream )
@@ -601,17 +539,63 @@ void sout_MuxSendBuffer( sout_mux_t *p_mux, sout_input_t *p_input,
         const int64_t i_caching = var_GetInteger( p_mux->p_sout, "sout-mux-caching" ) * INT64_C(1000);
 
         if( p_mux->i_add_stream_start < 0 )
-            p_mux->i_add_stream_start = p_buffer->i_dts;
+            p_mux->i_add_stream_start = i_dts;
 
-        /* Wait until we have enought data before muxing */
+        /* Wait until we have enough data before muxing */
         if( p_mux->i_add_stream_start < 0 ||
-            p_buffer->i_dts < p_mux->i_add_stream_start + i_caching )
-            return;
+            i_dts < p_mux->i_add_stream_start + i_caching )
+            return VLC_SUCCESS;
         p_mux->b_waiting_stream = false;
     }
-    p_mux->pf_mux( p_mux );
+    return p_mux->pf_mux( p_mux );
 }
 
+void sout_MuxFlush( sout_mux_t *p_mux, sout_input_t *p_input )
+{
+    VLC_UNUSED(p_mux);
+    block_FifoEmpty( p_input->p_fifo );
+}
+
+/*****************************************************************************
+ * sout_MuxGetStream: find stream to be muxed
+ *****************************************************************************/
+int sout_MuxGetStream( sout_mux_t *p_mux, unsigned i_blocks, vlc_tick_t *pi_dts )
+{
+    vlc_tick_t i_dts = 0;
+    int     i_stream = -1;
+
+    assert( i_blocks > 0 );
+
+    for( int i = 0; i < p_mux->i_nb_inputs; i++ )
+    {
+        sout_input_t *p_input = p_mux->pp_inputs[i];
+        block_t *p_data;
+
+        if( block_FifoCount( p_input->p_fifo ) < i_blocks )
+        {
+            if( (!p_mux->b_add_stream_any_time) &&
+                (p_input->p_fmt->i_cat != SPU_ES ) )
+            {
+                return -1;
+            }
+            /* FIXME: SPU muxing */
+            continue;
+        }
+
+        p_data = block_FifoShow( p_input->p_fifo );
+        if( i_stream < 0 || p_data->i_dts < i_dts )
+        {
+            i_stream = i;
+            i_dts    = p_data->i_dts;
+        }
+    }
+
+    if( pi_dts ) *pi_dts = i_dts;
+
+    return i_stream;
+}
+
+
 /*****************************************************************************
  *
  *****************************************************************************/
@@ -642,12 +626,12 @@ static int mrl_Parse( mrl_t *p_mrl, const char *psz_mrl )
             psz_parser++;
         }
     }
-#if defined( WIN32 ) || defined( UNDER_CE )
+#if defined( _WIN32 ) || defined( __OS2__ )
     if( psz_parser - psz_dup == 1 )
     {
         /* msg_Warn( p_sout, "drive letter %c: found in source string",
                           *psz_dup ) ; */
-        psz_parser = "";
+        *psz_parser = '\0';
     }
 #endif
 
@@ -748,67 +732,19 @@ static void mrl_Clean( mrl_t *p_mrl )
  ****************************************************************************
  ****************************************************************************/
 
-/* create a complete chain */
-/* chain format:
-    module{option=*:option=*}[:module{option=*:...}]
- */
-
-/*
- * parse module{options=str, option="str "}:
- *  return a pointer on the rest
- *  XXX: psz_chain is modified
- */
-
-/*
- * XXX name and p_cfg are used (-> do NOT free them)
- */
-sout_stream_t *sout_StreamNew( sout_instance_t *p_sout, char *psz_chain )
+/* Destroy a "stream_out" module */
+static void sout_StreamDelete( sout_stream_t *p_stream )
 {
-    static const char typename[] = "stream out";
-    sout_stream_t *p_stream;
+    sout_instance_t *p_sout = (sout_instance_t *)(p_stream->obj.parent);
 
-    if( !psz_chain )
-    {
-        msg_Err( p_sout, "invalid chain" );
-        return NULL;
-    }
-
-    p_stream = vlc_custom_create( p_sout, sizeof( *p_stream ),
-                                  VLC_OBJECT_GENERIC, typename );
-    if( !p_stream )
-        return NULL;
-
-    p_stream->p_sout   = p_sout;
-    p_stream->p_sys    = NULL;
-
-    p_stream->psz_next =
-        config_ChainCreate( &p_stream->psz_name, &p_stream->p_cfg, psz_chain);
-
-    msg_Dbg( p_sout, "stream=`%s'", p_stream->psz_name );
-
-    vlc_object_attach( p_stream, p_sout );
-
-    p_stream->p_module =
-        module_need( p_stream, "sout stream", p_stream->psz_name, true );
-
-    if( !p_stream->p_module )
-    {
-        sout_StreamDelete( p_stream );
-        return NULL;
-    }
-
-    return p_stream;
-}
-
-void sout_StreamDelete( sout_stream_t *p_stream )
-{
     msg_Dbg( p_stream, "destroying chain... (name=%s)", p_stream->psz_name );
 
-    vlc_object_detach( p_stream );
-    if( p_stream->p_module ) module_unneed( p_stream, p_stream->p_module );
+    p_sout->i_out_pace_nocontrol -= p_stream->pace_nocontrol;
+
+    if( p_stream->p_module != NULL )
+        module_unneed( p_stream, p_stream->p_module );
 
     FREENULL( p_stream->psz_name );
-    FREENULL( p_stream->psz_next );
 
     config_ChainDestroy( p_stream->p_cfg );
 
@@ -816,8 +752,158 @@ void sout_StreamDelete( sout_stream_t *p_stream )
     vlc_object_release( p_stream );
 }
 
-static char *_sout_stream_url_to_chain( vlc_object_t *p_this,
-                                        const char *psz_url )
+/* Destroy a "stream_out" modules chain
+ *
+ * p_first is the first module to be destroyed in the chain
+ * p_last is the last module to be destroyed
+ *  if NULL, all modules are destroyed
+ *  if not NULL, modules following it must be destroyed separately
+ */
+void sout_StreamChainDelete(sout_stream_t *p_first, sout_stream_t *p_last)
+{
+    while(p_first != NULL)
+    {
+        sout_stream_t *p_next = p_first->p_next;
+
+        sout_StreamDelete(p_first);
+        if(p_first == p_last)
+           break;
+        p_first = p_next;
+    }
+}
+
+/* Create a "stream_out" module, which may forward its ES to p_next module */
+/*
+ * XXX name and p_cfg are used (-> do NOT free them)
+ */
+static sout_stream_t *sout_StreamNew( sout_instance_t *p_sout, char *psz_name,
+                               config_chain_t *p_cfg, sout_stream_t *p_next)
+{
+    sout_stream_t *p_stream;
+
+    assert(psz_name);
+
+    p_stream = vlc_custom_create( p_sout, sizeof( *p_stream ), "stream out" );
+    if( !p_stream )
+        return NULL;
+
+    p_stream->p_sout   = p_sout;
+    p_stream->psz_name = psz_name;
+    p_stream->p_cfg    = p_cfg;
+    p_stream->p_next   = p_next;
+    p_stream->pf_flush = NULL;
+    p_stream->pf_control = NULL;
+    p_stream->pace_nocontrol = false;
+    p_stream->p_sys = NULL;
+
+    msg_Dbg( p_sout, "stream=`%s'", p_stream->psz_name );
+
+    p_stream->p_module =
+        module_need( p_stream, "sout stream", p_stream->psz_name, true );
+
+    if( !p_stream->p_module )
+    {
+        /* those must be freed by the caller if creation failed */
+        p_stream->psz_name = NULL;
+        p_stream->p_cfg = NULL;
+
+        sout_StreamDelete( p_stream );
+        return NULL;
+    }
+
+    p_sout->i_out_pace_nocontrol += p_stream->pace_nocontrol;
+    return p_stream;
+}
+
+/* Creates a complete "stream_out" modules chain
+ *
+ *  chain format: module1{option=*:option=*}[:module2{option=*:...}]
+ *
+ *  The modules are created starting from the last one and linked together
+ *  A pointer to the last module created is stored if pp_last isn't NULL, to
+ *  make sure sout_StreamChainDelete doesn't delete modules created in another
+ *  place.
+ *
+ *  Returns a pointer to the first module.
+ */
+sout_stream_t *sout_StreamChainNew(sout_instance_t *p_sout, const char *psz_chain,
+                                sout_stream_t *p_next, sout_stream_t **pp_last)
+{
+    if(!psz_chain || !*psz_chain)
+    {
+        if(pp_last) *pp_last = NULL;
+        return p_next;
+    }
+
+    char *psz_parser = strdup(psz_chain);
+    if(!psz_parser)
+        return NULL;
+
+    vlc_array_t cfg, name;
+    vlc_array_init(&cfg);
+    vlc_array_init(&name);
+
+    /* parse chain */
+    while(psz_parser)
+    {
+        config_chain_t *p_cfg;
+        char *psz_name;
+        char *psz_rest_chain = config_ChainCreate( &psz_name, &p_cfg, psz_parser );
+        free( psz_parser );
+        psz_parser = psz_rest_chain;
+
+        vlc_array_append_or_abort(&cfg, p_cfg);
+        vlc_array_append_or_abort(&name, psz_name);
+    }
+
+    size_t i = vlc_array_count(&name);
+    vlc_array_t module;
+    vlc_array_init(&module);
+    while(i--)
+    {
+        p_next = sout_StreamNew( p_sout, vlc_array_item_at_index(&name, i),
+            vlc_array_item_at_index(&cfg, i), p_next);
+
+        if(!p_next)
+            goto error;
+
+        if(i == vlc_array_count(&name) - 1 && pp_last)
+            *pp_last = p_next;   /* last module created in the chain */
+
+        vlc_array_append_or_abort(&module, p_next);
+    }
+
+    vlc_array_clear(&name);
+    vlc_array_clear(&cfg);
+    vlc_array_clear(&module);
+
+    return p_next;
+
+error:
+
+    i++;    /* last module couldn't be created */
+
+    /* destroy all modules created, starting with the last one */
+    int modules = vlc_array_count(&module);
+    while(modules--)
+        sout_StreamDelete(vlc_array_item_at_index(&module, modules));
+    vlc_array_clear(&module);
+
+    /* then destroy all names and config which weren't destroyed by
+     * sout_StreamDelete */
+    while(i--)
+    {
+        free(vlc_array_item_at_index(&name, i));
+        config_ChainDestroy(vlc_array_item_at_index(&cfg, i));
+    }
+    vlc_array_clear(&name);
+    vlc_array_clear(&cfg);
+
+    return NULL;
+}
+
+static char *sout_stream_url_to_chain( bool b_sout_display,
+                                       const char *psz_url )
 {
     mrl_t       mrl;
     char        *psz_chain;
@@ -863,10 +949,10 @@ rtp:
     }
 
     /* Duplicate and wrap if sout-display is on */
-    if (psz_chain && (config_GetInt( p_this, "sout-display" ) > 0))
+    if (psz_chain && b_sout_display)
     {
         char *tmp;
-        if (asprintf (&tmp, "duplicate{dst=display,dst=%s}", tmp) == -1)
+        if (asprintf (&tmp, "duplicate{dst=display,dst=%s}", psz_chain) == -1)
             tmp = NULL;
         free (psz_chain);
         psz_chain = tmp;
@@ -879,7 +965,5 @@ rtp:
 #undef sout_EncoderCreate
 encoder_t *sout_EncoderCreate( vlc_object_t *p_this )
 {
-    static const char type[] = "encoder";
-    return vlc_custom_create( p_this, sizeof( encoder_t ), VLC_OBJECT_GENERIC,
-                              type );
+    return vlc_custom_create( p_this, sizeof( encoder_t ), "encoder" );
 }
